/* The copyright in this software is beinOMg made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2013, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPrediction.cpp
    \brief    prediction class
*/

#include <memory.h>
#include "TComPrediction.h"
#include "TComTU.h"

#if SWP_M0052
  #define MAX_SAD 255*4
  #define GET_SAD_PS0(A1, A3, B1, B3)                   (abs(A1-B1) + abs(A3-B3))

#define MAX(A,B) A>B?A:B
#define MIN(A,B) A<B?A:B

#endif

//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Tables
// ====================================================================================================================

const UChar TComPrediction::m_aucIntraFilter[MAX_NUM_CHANNEL_TYPE][MAX_INTRA_FILTER_DEPTHS] =
{
  { // Luma
    10, //4x4
    7, //8x8
    1, //16x16
    0, //32x32
    10, //64x64
  },
  { // Chroma
    10, //4xn
    7, //8xn
    1, //16xn
    0, //32xn
    10, //64xn
  }

};

// ====================================================================================================================
// Constructor / destructor / initialize
// ====================================================================================================================

TComPrediction::TComPrediction()
: m_pLumaRecBuffer(0)
, m_iLumaRecStride(0)
{
  for(UInt ch=0; ch<MAX_NUM_COMPONENT; ch++)
  {
    for(UInt buf=0; buf<2; buf++)
    {
      m_piYuvExt[ch][buf] = NULL;
    }
  }
}

#if SWP_M0052
Void TComPrediction::initSWP(TComSPS* pSPS)
{
        UInt W_Luma_SWP[105] = {2097152, 1812405, 1566320, 1353648, 1169852, 1011012, 873739, 755104, 652577, 563972, 487397, 421219, 364027, 314600, 271884, 234968, 203065,
                                                        175493, 151665, 131072, 113275, 97895, 84603, 73116, 63188, 54609, 47194, 40786, 35248, 30462, 26326, 22752, 19662, 16993, 14686, 12692, 10968,
                                                        9479, 8192, 7080, 6118, 5288, 4570, 3949, 3413, 2950, 2549, 2203, 1904, 1645, 1422, 1229, 1062, 918, 793, 686, 592, 512, 442, 382, 330, 286, 247,
                                                        213, 184, 159, 138, 119, 103, 89, 77, 66, 57, 50, 43, 37, 32, 28, 24, 21, 18, 15, 13, 12, 10, 9, 7, 6, 6, 5, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1
                                                        };
        UInt W_Chroma422[78] = {2097152, 1720368, 1411279, 1157722, 949720, 779089, 639114, 524288, 430092, 352820, 289430, 237430, 194772, 159779, 131072, 107523, 88205, 72358,
                                                        59358, 48693, 39945, 32768, 26881, 22051, 18089, 14839, 12173, 9986, 8192, 6720, 5513, 4522, 3710, 3043, 2497, 2048, 1680, 1378, 1131, 927, 761,
                                                        624, 512, 420, 345, 283, 232, 190, 156, 128, 105, 86, 71, 58, 48, 39, 32, 26, 22, 18, 14, 12, 10, 8, 7, 5, 4, 4, 3, 2, 2, 2, 1, 1, 1, 1, 1, 1
                                                        };
        UInt W_Chroma420[50] = {2097152, 1541128, 1132525, 832255, 611597, 449443, 330281, 242713, 178361, 131072, 96321, 70783, 52016, 38225, 28090, 20643, 15170, 11148, 8192,
                                                        6020, 4424, 3251, 2389, 1756, 1290, 948, 697, 512, 376, 276, 203, 149, 110, 81, 59, 44, 32, 24, 17, 13, 9, 7, 5, 4, 3, 2, 1, 1, 1, 1
                                                        };

        m_xSizeSWP = MAX_CU_SIZE;
        m_ySizeSWP = MAX_CU_SIZE;
        m_leftOffsetSWP = 2;
        m_topOffsetSWP = 2;
        m_StrideSWP = m_xSizeSWP + m_leftOffsetSWP;

        assert(!m_pBufferSWP);
        assert(!m_puiNLMTableSADLuma);
        assert(!m_puiNLMTableSADChroma);

        m_pBufferSWP = new Pel[(m_xSizeSWP+m_leftOffsetSWP)*(m_ySizeSWP + m_topOffsetSWP)]();
        UInt m_ChromaFormatIDC = pSPS->getChromaFormatIdc();

        m_BitShiftSWPLuma       = pSPS->getBitDepth(CHANNEL_TYPE_LUMA) - 8;
        m_BitShiftSWPChroma     = pSPS->getBitDepth(CHANNEL_TYPE_CHROMA) - 8;
        m_MaxValueLuma = (1<<pSPS->getBitDepth(CHANNEL_TYPE_LUMA))-1;
        m_MaxValueChroma = (1<<pSPS->getBitDepth(CHANNEL_TYPE_CHROMA))-1;

        m_puiNLMTableSADLuma = new UInt[MAX_SAD+1]();
        for(int i=0; i<105; i++)
                m_puiNLMTableSADLuma[i] = W_Luma_SWP[i]>>m_BitShiftSWPLuma;

        switch (m_ChromaFormatIDC)
        {
          case CHROMA_400:
                  break;
          case CHROMA_420:
                  m_puiNLMTableSADChroma = new UInt[MAX_SAD+1]();
                  for(int i=0; i<50; i++)
                          m_puiNLMTableSADChroma[i] = W_Chroma420[i]>>m_BitShiftSWPChroma;
                  break;
          case CHROMA_422:
                  m_puiNLMTableSADChroma = new UInt[MAX_SAD+1]();
                  for(int i=0; i<78; i++)
                          m_puiNLMTableSADChroma[i] = W_Chroma422[i]>>m_BitShiftSWPChroma;
                  break;
          case CHROMA_444:
                  m_puiNLMTableSADChroma = m_puiNLMTableSADLuma;
                  break;
          default:
                  std::cerr << "\n!!!!  Invalid chroma format!";
                  exit(1);
        }
}
#endif  // SWP

TComPrediction::~TComPrediction()
{
  for(UInt ch=0; ch<MAX_NUM_COMPONENT; ch++)
  {
    for(UInt buf=0; buf<NUM_PRED_BUF; buf++)
    {
      delete [] m_piYuvExt[ch][buf];
    }
  }

  for(UInt i=0; i<NUM_REF_PIC_LIST_01; i++)
  {
    m_acYuvPred[i].destroy();
  }

  m_cYuvPredTemp.destroy();

  if( m_pLumaRecBuffer )
  {
    delete [] m_pLumaRecBuffer;
  }
  
  for (UInt i = 0; i < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; i++)
  {
    for (UInt j = 0; j < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; j++)
    {
      m_filteredBlock[i][j].destroy();
    }
    m_filteredBlockTmp[i].destroy();
  }

#if SWP_M0052
  if (!m_pBufferSWP)
  {
    delete m_pBufferSWP;
    m_pBufferSWP = NULL;
  }
  if (!m_puiNLMTableSADChroma)
  {
    delete m_puiNLMTableSADChroma;
    m_puiNLMTableSADChroma = NULL;
  }
  if (!m_puiNLMTableSADLuma)
  {
    delete m_puiNLMTableSADLuma;
    m_puiNLMTableSADLuma = NULL;
  }
#endif
}

Void TComPrediction::initTempBuff(ChromaFormat chromaFormatIDC)
{
  if( m_piYuvExt[COMPONENT_Y][PRED_BUF_UNFILTERED] == NULL ) // check if first is null (in which case, nothing initialised yet)
  {
    Int extWidth  = MAX_CU_SIZE + 16; 
    Int extHeight = MAX_CU_SIZE + 1;

    for (UInt i = 0; i < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; i++)
    {
      m_filteredBlockTmp[i].create(extWidth, extHeight + 7, chromaFormatIDC);
      for (UInt j = 0; j < LUMA_INTERPOLATION_FILTER_SUB_SAMPLE_POSITIONS; j++)
      {
        m_filteredBlock[i][j].create(extWidth, extHeight, chromaFormatIDC);
      }
    }

    m_iYuvExtSize = (MAX_CU_SIZE*2+1) * (MAX_CU_SIZE*2+1);
    for(UInt ch=0; ch<MAX_NUM_COMPONENT; ch++)
    {
      for(UInt buf=0; buf<NUM_PRED_BUF; buf++)
      {
        m_piYuvExt[ch][buf] = new Pel[ m_iYuvExtSize ];
      }
    }

    // new structure
    for(UInt i=0; i<NUM_REF_PIC_LIST_01; i++)
    {
      m_acYuvPred[i] .create( MAX_CU_SIZE, MAX_CU_SIZE, chromaFormatIDC );
    }

    m_cYuvPredTemp.create( MAX_CU_SIZE, MAX_CU_SIZE, chromaFormatIDC );
  }


  if (m_iLumaRecStride != (MAX_CU_SIZE>>1) + 1)
  {
    m_iLumaRecStride =  (MAX_CU_SIZE>>1) + 1;
    if (!m_pLumaRecBuffer)
    {
      m_pLumaRecBuffer = new Pel[ m_iLumaRecStride * m_iLumaRecStride ];
    }
  }
}

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

// Function for calculating DC value of the reference samples used in Intra prediction
//NOTE: Bit-Limit - 25-bit source
Pel TComPrediction::predIntraGetPredValDC( const Pel* pSrc, Int iSrcStride, UInt iWidth, UInt iHeight, ChannelType channelType, ChromaFormat format, Bool bAbove, Bool bLeft )
{
  Int iInd, iSum = 0;
  Pel pDcVal;

  if (bAbove)
  {
    for (iInd = 0;iInd < iWidth;iInd++)
    {
      iSum += pSrc[iInd-iSrcStride];
    }
  }
  if (bLeft)
  {
    for (iInd = 0;iInd < iHeight;iInd++)
    {
      iSum += pSrc[iInd*iSrcStride-1];
    }
  }

  if (bAbove && bLeft)
  {
    pDcVal = (iSum + iWidth) / (iWidth + iHeight);
  }
  else if (bAbove)
  {
    pDcVal = (iSum + iWidth/2) / iWidth;
  }
  else if (bLeft)
  {
    pDcVal = (iSum + iHeight/2) / iHeight;
  }
  else
  {
    pDcVal = pSrc[-1]; // Default DC value already calculated and placed in the prediction array if no neighbors are available
  }
  
  return pDcVal;
}

// Function for deriving the angular Intra predictions

/** Function for deriving the simplified angular intra predictions.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 * \param dirMode the intra prediction mode index
 * \param blkAboveAvailable boolean indication if the block above is available
 * \param blkLeftAvailable boolean indication if the block to the left is available
 *
 * This function derives the prediction samples for the angular mode based on the prediction direction indicated by
 * the prediction mode index. The prediction direction is given by the displacement of the bottom row of the block and
 * the reference row above the block in the case of vertical prediction or displacement of the rightmost column
 * of the block and reference column left from the block in the case of the horizontal prediction. The displacement
 * is signalled at 1/32 pixel accuracy. When projection of the predicted pixel falls inbetween reference samples,
 * the predicted value for the pixel is linearly interpolated from the reference samples. All reference samples are taken
 * from the extended main reference.
 */
//NOTE: Bit-Limit - 25-bit source
Void TComPrediction::xPredIntraAng(       Int bitDepth,
                                    const Pel* pSrc,     Int srcStride,
                                          Pel* pTrueDst, Int dstStrideTrue,
                                          UInt uiWidth, UInt uiHeight, ChannelType channelType, ChromaFormat format,
                                          UInt dirMode, Bool blkAboveAvailable, Bool blkLeftAvailable )
{
  Int width=Int(uiWidth);
  Int height=Int(uiHeight);

  // Map the mode index to main prediction direction and angle
  assert( dirMode != PLANAR_IDX ); //no planar
  const Bool modeDC        = dirMode==DC_IDX;

  // Do the DC prediction
  if (modeDC)
  {
    const Pel dcval = predIntraGetPredValDC(pSrc, srcStride, width, height, channelType, format, blkAboveAvailable, blkLeftAvailable);

    for (Int y=height;y>0;y--, pTrueDst+=dstStrideTrue)
    {
      for (Int x=0; x<width;) // width is always a multiple of 4.
      {
        pTrueDst[x++] = dcval;
      }
    }
  }
  else // Do angular predictions
  {
    const Bool       bIsModeVer         = (dirMode >= 18);
    const Int        intraPredAngleMode = (bIsModeVer) ? (Int)dirMode - VER_IDX :  -((Int)dirMode - HOR_IDX);
    const Int        absAngMode         = abs(intraPredAngleMode);
    const Int        signAng            = intraPredAngleMode < 0 ? -1 : 1;
    const Bool       edgeFilter         = isLuma(channelType) && (width <= MAXIMUM_INTRA_FILTERED_WIDTH) && (height <= MAXIMUM_INTRA_FILTERED_HEIGHT);

    // Set bitshifts and scale the angle parameter to block size
    static const Int angTable[9]    = {0,    2,    5,   9,  13,  17,  21,  26,  32};
    static const Int invAngTable[9] = {0, 4096, 1638, 910, 630, 482, 390, 315, 256}; // (256 * 32) / Angle
    Int invAngle                    = invAngTable[absAngMode];
    Int absAng                      = angTable[absAngMode];
    Int intraPredAngle              = signAng * absAng;

#if !RExt__M0127_CHROMA_422_INTRA_ANGLE_MAPPING
    if ((channelType == CHANNEL_TYPE_CHROMA) && (format == CHROMA_422))
    {
      intraPredAngle = bIsModeVer ? (intraPredAngle>>1) : 2*intraPredAngle;
      invAngle       = bIsModeVer ? 2*invAngle          : (invAngle>>1);

#if (RExt__SQUARE_TRANSFORM_CHROMA_422 != 0)
      if (std::abs(intraPredAngle) > 32)
      {
        intraPredAngle = 32 * signAng;
        invAngle       = 256;
      }
#endif
    }
#endif

    Pel* refMain;
    Pel* refSide;

    Pel  refAbove[2*MAX_CU_SIZE+1];
    Pel  refLeft[2*MAX_CU_SIZE+1];
    
    // Initialise the Main and Left reference array.
    if (intraPredAngle < 0)
    {
      const Int refMainOffsetPreScale = (bIsModeVer ? height : width ) - 1;
      const Int refMainOffset         = height - 1;
      for (Int x=0;x<width+1;x++)
      {
        refAbove[x+refMainOffset] = pSrc[x-srcStride-1];
      }
      for (Int y=0;y<height+1;y++)
      {
        refLeft[y+refMainOffset] = pSrc[(y-1)*srcStride-1];
      }
      refMain = (bIsModeVer ? refAbove : refLeft)  + refMainOffset;
      refSide = (bIsModeVer ? refLeft  : refAbove) + refMainOffset;

      // Extend the Main reference to the left.
      Int invAngleSum    = 128;       // rounding for (shift by 8)
      for (Int k=-1; k>(refMainOffsetPreScale+1)*intraPredAngle>>5; k--)
      {
        invAngleSum += invAngle;
        refMain[k] = refSide[invAngleSum>>8];
      }
    }
    else
    {
      for (Int x=0;x<2*width+1;x++)
      {
        refAbove[x] = pSrc[x-srcStride-1];
      }
      for (Int y=0;y<2*height+1;y++)
      {
        refLeft[y] = pSrc[(y-1)*srcStride-1];
      }
      refMain = bIsModeVer ? refAbove : refLeft ;
      refSide = bIsModeVer ? refLeft  : refAbove;
    }

    // swap width/height if we are doing a horizontal mode:
    Pel tempArray[MAX_CU_SIZE*MAX_CU_SIZE];
    const Int dstStride = bIsModeVer ? dstStrideTrue : MAX_CU_SIZE;
    Pel *pDst = bIsModeVer ? pTrueDst : tempArray;
    if (!bIsModeVer)
    {
      std::swap(width, height);
    }

    if (intraPredAngle == 0)  // pure vertical or pure horizontal
    {
      for (Int y=0;y<height;y++)
      {
        for (Int x=0;x<width;x++)
        {
          pDst[y*dstStride+x] = refMain[x+1];
        }
      }

      if (edgeFilter)
      {
        for (Int y=0;y<height;y++)
        {
          pDst[y*dstStride] = Clip3 (0, ((1 << bitDepth) - 1), pDst[y*dstStride] + (( refSide[y+1] - refSide[0] ) >> 1) );
        }
      }
    }
    else
    {
      Pel *pDsty=pDst;

      for (Int y=0, deltaPos=intraPredAngle; y<height; y++, deltaPos+=intraPredAngle, pDsty+=dstStride)
      {
        const Int deltaInt   = deltaPos >> 5;
        const Int deltaFract = deltaPos & (32 - 1);

        if (deltaFract)
        {
          // Do linear filtering
          const Pel *pRM=refMain+deltaInt+1;
          Int lastRefMainPel=*pRM++;
          for (Int x=0;x<width;pRM++,x++)
          {
            Int thisRefMainPel=*pRM;
            pDsty[x+0] = (Pel) ( ((32-deltaFract)*lastRefMainPel + deltaFract*thisRefMainPel +16) >> 5 );
            lastRefMainPel=thisRefMainPel;
          }
        }
        else
        {
          // Just copy the integer samples
          for (Int x=0;x<width; x++)
          {
            pDsty[x] = refMain[x+deltaInt+1];
          }
        }
      }
    }

    // Flip the block if this is the horizontal mode
    if (!bIsModeVer)
    {
      for (Int y=0; y<height; y++)
      {
        for (Int x=0; x<width; x++)
        {
          pTrueDst[x*dstStrideTrue] = pDst[x];
        }
        pTrueDst++;
        pDst+=dstStride;
      }
    }
  }
}


#if RExt__M0056_SAMPLE_ADAPTIVE_INTRA_PREDICT
Void TComPrediction::predIntraAng( const ComponentID compID, UInt uiDirMode, Pel* piOrg /* Will be null for decoding */, UInt uiOrgStride, Pel* piPred, UInt uiStride, TComTU &rTu, Bool bAbove, Bool bLeft, const Bool bUseFilteredPredSamples )
#else
Void TComPrediction::predIntraAng( const ComponentID compID, UInt uiDirMode, Pel* piPred, UInt uiStride, TComTU &rTu, Bool bAbove, Bool bLeft, const Bool bUseFilteredPredSamples )
#endif
{
  const ChromaFormat   format      = rTu.GetChromaFormat();
  const ChannelType    channelType = toChannelType(compID);
  const TComRectangle &rect        = rTu.getRect(isLuma(compID) ? COMPONENT_Y : COMPONENT_Cb);
  const Int            iWidth      = rect.width;
  const Int            iHeight     = rect.height;

  assert( g_aucConvertToBit[ iWidth ] >= 0 ); //   4x  4
  assert( g_aucConvertToBit[ iWidth ] <= 5 ); // 128x128
  //assert( iWidth == iHeight  );

        Pel *pDst = piPred;
#if !RExt__M0056_SAMPLE_ADAPTIVE_INTRA_PREDICT
  const Pel *ptrSrc = getPredictorPtr( compID, bUseFilteredPredSamples ) ;
#endif

  // get starting pixel in block
  const Int sw = (2 * iWidth + 1);

#if RExt__M0056_SAMPLE_ADAPTIVE_INTRA_PREDICT
  if ( UseSampleAdaptiveIntraPrediction(rTu, uiDirMode) )
  {
    const Pel *ptrSrc = getPredictorPtr( compID, false ) ;
    // Sample Adaptive intra-Prediction (SAP)
    if (uiDirMode==HOR_IDX)
    {
      // left column filled with reference samples
      // remaining columns filled with piOrg data (if available).
      for(Int y=0; y<iHeight; y++)
      {
        piPred[y*uiStride+0] = ptrSrc[(y+1)*sw];
      }
      if (piOrg!=0)
      {
        piPred+=1; // miss off first column
        for(Int y=0; y<iHeight; y++, piPred+=uiStride, piOrg+=uiOrgStride)
        {
          memcpy(piPred, piOrg, (iWidth-1)*sizeof(Pel));
        }
      }
    }
    else // VER_IDX
    {
      // top row filled with reference samples
      // remaining rows filled with piOrd data (if available)
      for(Int x=0; x<iWidth; x++)
      {
        piPred[x] = ptrSrc[x+1];
      }
      if (piOrg!=0)
      {
        piPred+=uiStride; // miss off the first row
        for(Int y=1; y<iHeight; y++, piPred+=uiStride, piOrg+=uiOrgStride)
        {
          memcpy(piPred, piOrg, iWidth*sizeof(Pel));
        }
      }
    }
  }
  else
  {
    const Pel *ptrSrc = getPredictorPtr( compID, bUseFilteredPredSamples ) ;
#endif
    if ( uiDirMode == PLANAR_IDX )
    {
      xPredIntraPlanar( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType, format );
    }
    else
    {
      // Create the prediction
      xPredIntraAng( g_bitDepth[channelType], ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType, format, uiDirMode, bAbove, bLeft );

      if(( uiDirMode == DC_IDX ) && bAbove && bLeft )
      {
        xDCPredFiltering( ptrSrc+sw+1, sw, pDst, uiStride, iWidth, iHeight, channelType );
      }
    }
#if RExt__M0056_SAMPLE_ADAPTIVE_INTRA_PREDICT
  }
#endif

}

/** Function for checking identical motion.
 * \param TComDataCU* pcCU
 * \param UInt PartAddr
 */
Bool TComPrediction::xCheckIdenticalMotion ( TComDataCU* pcCU, UInt PartAddr )
{
  if( pcCU->getSlice()->isInterB() && !pcCU->getSlice()->getPPS()->getWPBiPred() )
  {
    if( pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr) >= 0 && pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr) >= 0)
    {
      Int RefPOCL0 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_0, pcCU->getCUMvField(REF_PIC_LIST_0)->getRefIdx(PartAddr))->getPOC();
      Int RefPOCL1 = pcCU->getSlice()->getRefPic(REF_PIC_LIST_1, pcCU->getCUMvField(REF_PIC_LIST_1)->getRefIdx(PartAddr))->getPOC();
      if(RefPOCL0 == RefPOCL1 && pcCU->getCUMvField(REF_PIC_LIST_0)->getMv(PartAddr) == pcCU->getCUMvField(REF_PIC_LIST_1)->getMv(PartAddr))
      {
        return true;
      }
    }
  }
  return false;
}


#if SWP_M0052

inline Pel getBest_PS0_NS1(Pel* piSrc, Int srcStride)
{
  Int distA, distB, distC;
  Int A,B, C, AA,AC, BA,BC, CA,CC;
  Int i=0; Int j=0;

  A = piSrc[srcStride*(j) + i-1];     B = piSrc[srcStride*(j-1) + i-1],       C = piSrc[srcStride*(j-1) + i];
  AA = piSrc[srcStride*(j) + i-2];    AC = piSrc[srcStride*(j-1) + i-1];      distA = GET_SAD_PS0(A, C, AA, AC);
  BA = piSrc[srcStride*(j-1) + i-2];  BC = piSrc[srcStride*(j-2) + i-1];      distB = GET_SAD_PS0(A, C, BA, BC);
  CA = piSrc[srcStride*(j-1) + i-1];  CC = piSrc[srcStride*(j-2) + i];        distC = GET_SAD_PS0(A, C, CA, CC);

  // C->D->A->B
  if (distC < distB)
    {
      if (distC < distA)
        return C;
      else
        return A;
    }
  else
    {
      if (distB < distA)
        return B;
      else
        return A;
    }
}

inline Pel getBest_PS0_NS2(Pel* piSrc, Int srcStride)
{
  Int distA, distB, distC, distD;
  Int A,B,C,D, AA,AC, BA,BC, CA,CC, DA,DC;
  Int i=0; Int j=0;


  A = piSrc[srcStride*(j) + i-1];     B = piSrc[srcStride*(j-1) + i-1],       C = piSrc[srcStride*(j-1) + i],       D = piSrc[srcStride*(j-1) + i+1];
  AA = piSrc[srcStride*(j) + i-2];    AC = piSrc[srcStride*(j-1) + i-1];      distA = GET_SAD_PS0(A, C, AA, AC);
  BA = piSrc[srcStride*(j-1) + i-2];  BC = piSrc[srcStride*(j-2) + i-1];      distB = GET_SAD_PS0(A, C, BA, BC);
  CA = piSrc[srcStride*(j-1) + i-1];  CC = piSrc[srcStride*(j-2) + i];        distC = GET_SAD_PS0(A, C, CA, CC);
  DA = piSrc[srcStride*(j-1) + i];    DC = piSrc[srcStride*(j-2) + i+1];      distD = GET_SAD_PS0(A, C, DA, DC);
  Int dist1, dist2, val1, val2;

  // C->D->A->B
    if (distB < distA)
      {
        dist1 = distB; val1 = B;
      }
    else
      {
        dist1 = distA; val1 = A;
      }

    if (distD < distC)
      {
        dist2 = distD; val2 = D;
      }
    else
      {
        dist2 = distC; val2 = C;
      }

  if (dist1 < dist2)
    return (Pel)val1;
  else
    return (Pel)val2;
}


Void TComPrediction::xPredBlkNLM(Pel* piSrc, Int srcStride, Pel* piOrg, Pel* piPred, UInt dstStride, UInt iWidth, Int iHeight, Bool bAbove, Bool bLeft, bool isChromaFlag)
{
  int i;

    {

  if (bAbove && bLeft)
      for ( int j=0; j<iHeight; j++)
      {
        for (i=0; i<iWidth-1; i++)
        {
            piPred[dstStride*j + i] = (Pel)getBest_PS0_NS2(&(piSrc[srcStride*j + i]), srcStride);
            piSrc[srcStride*j + i] = piOrg[dstStride*j + i];
        }
        piPred[dstStride*j + i] = (Pel)getBest_PS0_NS1(&(piSrc[srcStride*j + i]), srcStride);
        piSrc[srcStride*j + i] = piOrg[dstStride*j + i];
      }

    }
}

Void TComPrediction::xRecoBlkNLM(Pel* piSrc, Int srcStride, Pel* piResi, Pel* piPred, UInt dstStride, UInt iWidth, Int iHeight, Bool bAbove, Bool bLeft, bool isChromaFlag)
{
  int i;


      {
  if (bAbove && bLeft)
      for ( int j=0; j<iHeight; j++)
      {
        for (i=0; i<iWidth-1; i++)
        {
            piPred[dstStride*j + i] = (Pel)getBest_PS0_NS2(&(piSrc[srcStride*j + i]), srcStride);
            piSrc[srcStride*j + i] = piPred[dstStride*j + i] + piResi[dstStride*j + i];
        }
        piPred[dstStride*j + i] = (Pel)getBest_PS0_NS1(&(piSrc[srcStride*j + i]), srcStride);
        piSrc[srcStride*j + i] = piPred[dstStride*j + i] + piResi[dstStride*j + i];
      }
      }
}

Void TComPrediction::initLeftAndTopSamplesSWP(TComTU& rTu, UInt left_offset, UInt top_offset, ComponentID compID)
{
  assert(left_offset <= m_leftOffsetSWP);
  assert(top_offset <= m_topOffsetSWP);

  Bool bLeftAvailable, bTopLeftAvailable, bTopAvailable;
  TComDataCU *pcCU=rTu.getCU();
  const UInt uiZorderIdxInPart=rTu.GetAbsPartIdxTU();
  const UInt widthTop        = rTu.getRect(compID).width;
  const UInt heightLeft       = rTu.getRect(compID).height;
  UInt escapeVal;

  Pel *piRecOrigin = pcCU->getPic()->getPicYuvRec()->getAddr(compID, pcCU->getAddr(), pcCU->getZorderIdxInCU()+uiZorderIdxInPart);
  UInt    recStride = pcCU->getPic()->getPicYuvRec()->getStride(compID);

  // check whether upper and left pixels are available?
  UInt tempPartUnitIdx;
  bTopLeftAvailable     = pcCU->getPUAboveLeft(tempPartUnitIdx, pcCU->getZorderIdxInCU()+(UInt)uiZorderIdxInPart)==NULL ? 0:1;
  bTopAvailable         = pcCU->getPUAbove(tempPartUnitIdx,     pcCU->getZorderIdxInCU()+(UInt)uiZorderIdxInPart)==NULL ? 0:1;
  bLeftAvailable        = pcCU->getPULeft(tempPartUnitIdx,      pcCU->getZorderIdxInCU()+(UInt)uiZorderIdxInPart)==NULL ? 0:1;

  if (compID == COMPONENT_Y)
          escapeVal = 1<<(7+m_BitShiftSWPLuma);
  else
          escapeVal = 1<<(7+m_BitShiftSWPChroma);

  // only top pixel
  Pel* pDst = m_pBufferSWP + left_offset;             // shifted to right
  if (bTopAvailable){
    Pel* pRec = piRecOrigin - recStride*top_offset;     // shifted to up
    for (int j=0; j<top_offset; j++){
      for(int i=0; i<widthTop; i++){
        pDst[i] = pRec[i];
      }
      pDst += m_StrideSWP;
      pRec += recStride;
    }
  }
  else{
    for (int j=0; j<top_offset; j++){
      for(int i=0; i<widthTop; i++){
        pDst[i] = escapeVal;
      }
      pDst += m_StrideSWP;
    }
  }

  // only left pixels
  pDst = m_pBufferSWP + top_offset*m_StrideSWP;  // shifted down
  if (bLeftAvailable){
    Pel* pRec = piRecOrigin - left_offset;              // shifted to left
    for (int j=0; j<heightLeft; j++){
      for(int i=0; i<left_offset; i++){
        pDst[i] = pRec[i];
      }
      pDst += m_StrideSWP;
      pRec += recStride;
    }
  }
  else{
    for (int j=0; j<heightLeft; j++){
      for(int i=0; i<left_offset; i++){
        pDst[i] = escapeVal;
      }
      pDst += m_StrideSWP;
    }
  }

  // only top left pixels
  pDst = m_pBufferSWP;                                           // not shifted
  if (bTopLeftAvailable){
    Pel* pRec = piRecOrigin - recStride*top_offset - left_offset;       // shifted to top left
    for (int j=0; j<top_offset; j++){
      for(int i=0; i<left_offset; i++){
        pDst[i] = pRec[i];
      }
      pDst += m_StrideSWP;
      pRec += recStride;
    }
  }
  else{
    for (int j=0; j<top_offset; j++){
      for(int i=0; i<left_offset; i++){
        pDst[i] = escapeVal;
      }
      pDst += m_StrideSWP;
    }
  }
}

#if SWP_DC_MODE
Void TComPrediction::predIntraSWP(Pel* piOrg, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Int left_offset, Int top_offset, ComponentID compID, UInt uiMode)
{
  if(compID == COMPONENT_Y)
    xPredBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piOrg, piPred, uiStride, iWidth, iHeight, true, true, false, uiMode);
  else
    xPredBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piOrg, piPred, uiStride, iWidth, iHeight, true, true, true, uiMode);
}

Void TComPrediction::recoIntraSWP(Pel* piResi, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Int left_offset, Int top_offset, ComponentID compID, UInt uiMode)
{
  if(compID == COMPONENT_Y)
    xRecoBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piResi, piPred, uiStride, iWidth, iHeight, true, true, false, uiMode);
  else
    xRecoBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piResi, piPred, uiStride, iWidth, iHeight, true, true, true, uiMode);
}
#else
Void TComPrediction::predIntraSWP(Pel* piOrg, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Int left_offset, Int top_offset, ComponentID compID)
{
  if(compID == COMPONENT_Y)
    xPredBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piOrg, piPred, uiStride, iWidth, iHeight, true, true, false);
  else
    xPredBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piOrg, piPred, uiStride, iWidth, iHeight, true, true, true);
}

Void TComPrediction::recoIntraSWP(Pel* piResi, Pel* piPred, UInt uiStride, Int iWidth, Int iHeight, Int left_offset, Int top_offset, ComponentID compID)
{
  if(compID == COMPONENT_Y)
    xRecoBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piResi, piPred, uiStride, iWidth, iHeight, true, true, false);
  else
    xRecoBlkNLM(m_pBufferSWP + top_offset*m_StrideSWP + left_offset, m_StrideSWP, piResi, piPred, uiStride, iWidth, iHeight, true, true, true);
}
#endif // SWP_DC_MODE
#endif //SWP

Void TComPrediction::motionCompensation ( TComDataCU* pcCU, TComYuv* pcYuvPred, RefPicList eRefPicList, Int iPartIdx )
{
  Int         iWidth;
  Int         iHeight;
  UInt        uiPartAddr;

  if ( iPartIdx >= 0 )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );
    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
    return;
  }

  for ( iPartIdx = 0; iPartIdx < pcCU->getNumPartInter(); iPartIdx++ )
  {
    pcCU->getPartIndexAndSize( iPartIdx, uiPartAddr, iWidth, iHeight );

    if ( eRefPicList != REF_PIC_LIST_X )
    {
      if( pcCU->getSlice()->getPPS()->getUseWP())
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred, true );
      }
      else
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
      if ( pcCU->getSlice()->getPPS()->getUseWP() )
      {
        xWeightedPredictionUni( pcCU, pcYuvPred, uiPartAddr, iWidth, iHeight, eRefPicList, pcYuvPred );
      }
    }
    else
    {
      if ( xCheckIdenticalMotion( pcCU, uiPartAddr ) )
      {
        xPredInterUni (pcCU, uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, pcYuvPred );
      }
      else
      {
        xPredInterBi  (pcCU, uiPartAddr, iWidth, iHeight, pcYuvPred );
      }
    }
  }
  return;
}

Void TComPrediction::xPredInterUni ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, RefPicList eRefPicList, TComYuv*& rpcYuvPred, Bool bi )
{
  Int         iRefIdx     = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );           assert (iRefIdx >= 0);
  TComMv      cMv         = pcCU->getCUMvField( eRefPicList )->getMv( uiPartAddr );
  pcCU->clipMv(cMv);

  for (UInt ch=COMPONENT_Y; ch<rpcYuvPred->getNumberValidComponents(); ch++)
    xPredInterBlk  (ComponentID(ch),  pcCU, pcCU->getSlice()->getRefPic( eRefPicList, iRefIdx )->getPicYuvRec(), uiPartAddr, &cMv, iWidth, iHeight, rpcYuvPred, bi );
}

Void TComPrediction::xPredInterBi ( TComDataCU* pcCU, UInt uiPartAddr, Int iWidth, Int iHeight, TComYuv*& rpcYuvPred )
{
  TComYuv* pcMbYuv;
  Int      iRefIdx[NUM_REF_PIC_LIST_01] = {-1, -1};

  for ( UInt refList = 0; refList < NUM_REF_PIC_LIST_01; refList++ )
  {
    RefPicList eRefPicList = (refList ? REF_PIC_LIST_1 : REF_PIC_LIST_0);
    iRefIdx[refList] = pcCU->getCUMvField( eRefPicList )->getRefIdx( uiPartAddr );

    if ( iRefIdx[refList] < 0 )
    {
      continue;
    }

    assert( iRefIdx[refList] < pcCU->getSlice()->getNumRefIdx(eRefPicList) );

    pcMbYuv = &m_acYuvPred[refList];
    if( pcCU->getCUMvField( REF_PIC_LIST_0 )->getRefIdx( uiPartAddr ) >= 0 && pcCU->getCUMvField( REF_PIC_LIST_1 )->getRefIdx( uiPartAddr ) >= 0 )
    {
      xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
    }
    else
    {
      if ( ( pcCU->getSlice()->getPPS()->getUseWP()       && pcCU->getSlice()->getSliceType() == P_SLICE ) || 
           ( pcCU->getSlice()->getPPS()->getWPBiPred()    && pcCU->getSlice()->getSliceType() == B_SLICE ) )
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv, true );
      }
      else
      {
        xPredInterUni ( pcCU, uiPartAddr, iWidth, iHeight, eRefPicList, pcMbYuv );
      }
    }
  }

  if ( pcCU->getSlice()->getPPS()->getWPBiPred()    && pcCU->getSlice()->getSliceType() == B_SLICE  )
  {
    xWeightedPredictionBi( pcCU, &m_acYuvPred[REF_PIC_LIST_0], &m_acYuvPred[REF_PIC_LIST_1], iRefIdx[REF_PIC_LIST_0], iRefIdx[REF_PIC_LIST_1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
  else if ( pcCU->getSlice()->getPPS()->getUseWP() && pcCU->getSlice()->getSliceType() == P_SLICE )
  {
    xWeightedPredictionUni( pcCU, &m_acYuvPred[REF_PIC_LIST_0], uiPartAddr, iWidth, iHeight, REF_PIC_LIST_0, rpcYuvPred );
  }
  else
  {
    xWeightedAverage( &m_acYuvPred[REF_PIC_LIST_0], &m_acYuvPred[REF_PIC_LIST_1], iRefIdx[REF_PIC_LIST_0], iRefIdx[REF_PIC_LIST_1], uiPartAddr, iWidth, iHeight, rpcYuvPred );
  }
}

/**
 * \brief Generate motion-compensated block
 *
 * \param cu       Pointer to current CU
 * \param refPic   Pointer to reference picture
 * \param partAddr Address of block within CU
 * \param mv       Motion vector
 * \param width    Width of block
 * \param height   Height of block
 * \param dstPic   Pointer to destination picture
 * \param bi       Flag indicating whether bipred is used
 */


Void TComPrediction::xPredInterBlk(const ComponentID compID, TComDataCU *cu, TComPicYuv *refPic, UInt partAddr, TComMv *mv, Int width, Int height, TComYuv *&dstPic, Bool bi )
{
  Int     refStride  = refPic->getStride(compID);
  Int     dstStride  = dstPic->getStride(compID);
  Int shiftHor=(2+refPic->getComponentScaleX(compID));
  Int shiftVer=(2+refPic->getComponentScaleY(compID));
  
  Int     refOffset  = (mv->getHor() >> shiftHor) + (mv->getVer() >> shiftVer) * refStride;
  
  Pel*    ref     = refPic->getAddr(compID, cu->getAddr(), cu->getZorderIdxInCU() + partAddr ) + refOffset;
  
  Pel*    dst = dstPic->getAddr( compID, partAddr );
  
  Int     xFrac  = mv->getHor() & ((1<<shiftHor)-1);
  Int     yFrac  = mv->getVer() & ((1<<shiftVer)-1);
  UInt    cxWidth  = width  >> refPic->getComponentScaleX(compID);
  UInt    cxHeight = height >> refPic->getComponentScaleY(compID);
  
  const ChromaFormat chFmt = cu->getPic()->getChromaFormat();
  
  if ( yFrac == 0 )
  {
    m_if.filterHor(compID, ref, refStride, dst,  dstStride, cxWidth, cxHeight, xFrac, !bi, chFmt);
  }
  else if ( xFrac == 0 )
  {
    m_if.filterVer(compID, ref, refStride, dst, dstStride, cxWidth, cxHeight, yFrac, true, !bi, chFmt);
  }
  else
  {
    Int   tmpStride = m_filteredBlockTmp[0].getStride(compID);
    Pel*  tmp       = m_filteredBlockTmp[0].getAddr(compID);

    const Int vFilterSize = isLuma(compID) ? NTAPS_LUMA : NTAPS_CHROMA;

    m_if.filterHor(compID, ref - ((vFilterSize>>1) -1)*refStride, refStride, tmp, tmpStride, cxWidth, cxHeight+vFilterSize-1, xFrac, false,      chFmt);
    m_if.filterVer(compID, tmp + ((vFilterSize>>1) -1)*tmpStride, tmpStride, dst, dstStride, cxWidth, cxHeight,               yFrac, false, !bi, chFmt);
  }
}

Void TComPrediction::xWeightedAverage( TComYuv* pcYuvSrc0, TComYuv* pcYuvSrc1, Int iRefIdx0, Int iRefIdx1, UInt uiPartIdx, Int iWidth, Int iHeight, TComYuv*& rpcYuvDst )
{
  if( iRefIdx0 >= 0 && iRefIdx1 >= 0 )
  {
    rpcYuvDst->addAvg( pcYuvSrc0, pcYuvSrc1, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 >= 0 && iRefIdx1 <  0 )
  {
    pcYuvSrc0->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
  else if ( iRefIdx0 <  0 && iRefIdx1 >= 0 )
  {
    pcYuvSrc1->copyPartToPartYuv( rpcYuvDst, uiPartIdx, iWidth, iHeight );
  }
}

// AMVP
Void TComPrediction::getMvPredAMVP( TComDataCU* pcCU, UInt uiPartIdx, UInt uiPartAddr, RefPicList eRefPicList, TComMv& rcMvPred )
{
  AMVPInfo* pcAMVPInfo = pcCU->getCUMvField(eRefPicList)->getAMVPInfo();

  if( pcAMVPInfo->iN <= 1 )
  {
    rcMvPred = pcAMVPInfo->m_acMvCand[0];

    pcCU->setMVPIdxSubParts( 0, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    pcCU->setMVPNumSubParts( pcAMVPInfo->iN, eRefPicList, uiPartAddr, uiPartIdx, pcCU->getDepth(uiPartAddr));
    return;
  }

  assert(pcCU->getMVPIdx(eRefPicList,uiPartAddr) >= 0);
  rcMvPred = pcAMVPInfo->m_acMvCand[pcCU->getMVPIdx(eRefPicList,uiPartAddr)];
  return;
}

/** Function for deriving planar intra prediction.
 * \param pSrc pointer to reconstructed sample array
 * \param srcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param dstStride the stride of the prediction sample array
 * \param width the width of the block
 * \param height the height of the block
 *
 * This function derives the prediction samples for planar mode (intra coding).
 */
//NOTE: Bit-Limit - 24-bit source
Void TComPrediction::xPredIntraPlanar( const Pel* pSrc, Int srcStride, Pel* rpDst, Int dstStride, UInt width, UInt height, ChannelType channelType, ChromaFormat format )
{
  assert(width <= height);

#if RExt__BACKWARDS_COMPATIBILITY_HM_TICKET_992
  Int leftColumn[MAX_CU_SIZE], topRow[MAX_CU_SIZE], bottomRow[MAX_CU_SIZE], rightColumn[MAX_CU_SIZE];
#else
  Int leftColumn[MAX_CU_SIZE+1], topRow[MAX_CU_SIZE+1], bottomRow[MAX_CU_SIZE], rightColumn[MAX_CU_SIZE];
#endif
  UInt shift1Dhor = g_aucConvertToBit[ width ] + 2;
  UInt shift1Dver = g_aucConvertToBit[ height ] + 2;

  // Get left and above reference column and row
  for(Int k=0;k<width+1;k++)
  {
    topRow[k] = pSrc[k-srcStride];
  }

  for (Int k=0; k < height+1; k++)
  {
    leftColumn[k] = pSrc[k*srcStride-1];
  }

  // Prepare intermediate variables used in interpolation
  Int bottomLeft = leftColumn[height];
  Int topRight   = topRow[width];

  for(Int k=0;k<width;k++)
  {
    bottomRow[k]  = bottomLeft - topRow[k];
    topRow[k]     <<= shift1Dver;
  }

  for(Int k=0;k<height;k++)
  {
    rightColumn[k]  = topRight - leftColumn[k];
    leftColumn[k]   <<= shift1Dhor;
  }

#if (RExt__SQUARE_TRANSFORM_CHROMA_422 != 0)
  const UInt topRowShift = 0;
#else
  const UInt topRowShift = (isChroma(channelType) && (format == CHROMA_422)) ? 1 : 0;
#endif

  // Generate prediction signal
  for (Int y=0;y<height;y++)
  {
    Int horPred = leftColumn[y] + width;
    for (Int x=0;x<width;x++)
    {
      horPred += rightColumn[y];
      topRow[x] += bottomRow[x];

      // NOTE: RExt - The intermediate shift right could be rolled into the final shift right,
      //              thereby increasing the accuracy of the calculation
      // eg  rpDst[y*dstStride+x] = ( (horPred<<topRowShift) + topRow[x] ) >> (shift1Dver+1);
      Int vertPred = ((topRow[x] + topRowShift)>>topRowShift);
      rpDst[y*dstStride+x] = ( horPred + vertPred ) >> (shift1Dhor+1);
    }
  }
}

/** Function for filtering intra DC predictor.
 * \param pSrc pointer to reconstructed sample array
 * \param iSrcStride the stride of the reconstructed sample array
 * \param rpDst reference to pointer for the prediction sample array
 * \param iDstStride the stride of the prediction sample array
 * \param iWidth the width of the block
 * \param iHeight the height of the block
 *
 * This function performs filtering left and top edges of the prediction samples for DC mode (intra coding).
 */
Void TComPrediction::xDCPredFiltering( const Pel* pSrc, Int iSrcStride, Pel*& rpDst, Int iDstStride, Int iWidth, Int iHeight, ChannelType channelType )
{
  Pel* pDst = rpDst;
  Int x, y, iDstStride2, iSrcStride2;

  if (isLuma(channelType) && (iWidth <= MAXIMUM_INTRA_FILTERED_WIDTH) && (iHeight <= MAXIMUM_INTRA_FILTERED_HEIGHT))
  {
    //top-left
    pDst[0] = (Pel)((pSrc[-iSrcStride] + pSrc[-1] + 2 * pDst[0] + 2) >> 2);

    //top row (vertical filter)
    for ( x = 1; x < iWidth; x++ )
    {
      pDst[x] = (Pel)((pSrc[x - iSrcStride] +  3 * pDst[x] + 2) >> 2);
    }

    //left column (horizontal filter)
    for ( y = 1, iDstStride2 = iDstStride, iSrcStride2 = iSrcStride-1; y < iHeight; y++, iDstStride2+=iDstStride, iSrcStride2+=iSrcStride )
    {
      pDst[iDstStride2] = (Pel)((pSrc[iSrcStride2] + 3 * pDst[iDstStride2] + 2) >> 2);
    }
  }

  return;
}

#if RExt__M0056_SAMPLE_ADAPTIVE_INTRA_PREDICT
/* Static member function */
Bool TComPrediction::UseSampleAdaptiveIntraPrediction(TComTU &rTu, const UInt uiDirMode)
{
  // TODO: RExt - possibly check other sub-layers profile idcs here?
  return (rTu.getCU()->getSlice()->getSPS()->getPTL()->getGeneralPTL()->getProfileIdc()==Profile::REXTDEV) &&
         rTu.getCU()->getCUTransquantBypass(rTu.GetAbsPartIdxTU()) &&
         (uiDirMode==HOR_IDX || uiDirMode==VER_IDX);
}
#endif

//! \}
